from pyrediseasyio.io.io_group import IOGroup
from pyrediseasyio.html.html_io_group import HMTLIOGroup
from pyrediseasyio.html.html_io import HTMLIO
from pyrediseasyio.io.single_io import SingleIO
from pyrediseasyio.io.boolean_io import BooleanIO
from pyrediseasyio.io.string_io import StringIO
from pyrediseasyio.io.float_io import FloatIO
from pyrediseasyio.io.integer_io import IntIO
from pyrediseasyio.io.trigger_io import TriggerIO



from pyrediseasyio.io.single_io import SingleIO


class TriggerIO(SingleIO):
    """ This IO does not actually write to the Redis server. Its intended use is to allow logic to be applied to its
    set or reset operations. For example, one might want to trigger a clearing of errors, in which case the set
    callback of the trigger would be wired to turning off several other IO"""

    def __init__(self, name: str, set_callback=None, read_callback=None, default = None, **kwargs):
        """
        :param name: A human readable name to be applied to the trigger
        :param set_callback: A method to call when setting the trigger, takes a single argument - value
        :param read_callback: A method to call when reading the trigger, takes no arguments, and return any object
        :param kwargs:
         - units:       str: Optional units to assign to the IO
         - on_value:    str: The 'display_value' property will optionally return this value when the value is True
         - on_value:    str: The 'display_value' property will optionally return this value when the value is False
        """
        super().__init__(name, addr=None, default=default, **kwargs)
        self.set_callback = set_callback
        self.read_callback = read_callback
        self._iter_index = 0

    def read(self):
        if self.read_callback:
            val = self.read_callback()
            return val
        return None

    def write(self, value):
        if self.set_callback:
            self.set_callback(value)

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        val = self.value
        if self._iter_index >= len(val):
            raise StopIteration
        else:
            self._iter_index += 1
            return val[self._iter_index - 1]




from pyrediseasyio.io.boolean_io import BooleanIO
from pyrediseasyio.io.io_group import IOGroup
from pyrediseasyio.html.html_io import HTMLIO
from typing import List, Callable
from dominate.tags import div, table, tr, th
import dominate
import json


class HMTLIOGroup:

    def __init__(self, io_group: IOGroup):
        self._io_group = io_group
        self.namespace = io_group.namespace


    def _heading(self, headers, row_tag: dominate.tags, cell_tag: dominate.tags):
        if headers:
            with row_tag(cls="easyio_header_row"):
                for h in headers:
                    cell_tag(h, cls="easyio_header_cell")


    def html(self, headers: List[str] = None, html_id: str = None, **kwargs):
        """
        :param headers: An optional list of headers that can be applied to the table
        :param html_id: An optional ID that can be given to the table
        :param kwargs: (by_names, by_type, by_lambda_each, by_lambda_results,
                        show_units, show_set, show_reset, set_text, reset_text)
        :return:    Returns a dominate table object that can be rendered to html.
        """
        attrs = self._io_group.get_attributes(**kwargs)
        html_id = f'{attrs[0].key}_io_container' if html_id is None else html_id

        with div(cls=f'easyio_container', id=html_id) as container:
            self._heading(headers, div, div)
            for attr in attrs:
                HTMLIO(attr).html(**kwargs)
        return container



    def html_table(self, headers: List[str] = None, html_id: str = None, **kwargs) -> table:
        """
        :param headers: An optional list of headers that can be applied to the table
        :param html_id: An optional ID that can be given to the table
        :param kwargs: (by_names, by_type, by_lambda_each, by_lambda_results,
                        show_units, show_set, show_reset, set_text, reset_text)
        :return:    Returns a dominate table object that can be rendered to html.
        """
        attrs = self._io_group.get_attributes(**kwargs)
        html_id = f'{attrs[0].addr}_io_container' if html_id is None else html_id

        with table(cls=f'easyio_container', id=html_id) as container:
            self._heading(headers, tr, th)
            for attr in attrs:
                HTMLIO(attr).html_row(**kwargs)
        return container



    def dumps(self, **kwargs):
        """
        Create a json string giving relevant information about each member of the IO Group useful at the web side.
        :param kwargs: (by_names, by_type, by_lambda_each, by_lambda_results)
        :return:
        """
        attrs = self._io_group.get_attributes(**kwargs)

        def f(a):
            io = HTMLIO(a)
            value, display_value = io.value, a.display_value
            return dict(id=io.value_id, name=io.name, value=io.value, units=io.units, display_value=display_value,
                        namespace=a.namespace, addr=a.addr)

        results = [f(a) for a in attrs]

        return json.dumps(results)





from flask import Flask, Response, jsonify
from pyrediseasyio import IOGroup, HMTLIOGroup, HTMLIO, TriggerIO
from pyrediseasyio import StringIO, BooleanIO, FloatIO, IntIO
from dominate.tags import head, script, body, link, div, table
import dominate
import random


app = Flask(__name__)


class TestGroup(IOGroup):
    FirstName = StringIO("First Name", default="Steve")
    LastName = StringIO("Last Name", default="Jackson")
    Age = IntIO("Age", default=48, units="years")
    IsProgrammer = BooleanIO("Is a programmer", on_value="Yes", off_value="No", default=True)
    IsFather = BooleanIO("Is a father", on_value="Definately", off_value="Not yet", default=True)
    BathroomBreaks = IntIO("Bathroom breaks", default=3)
    NameChange = TriggerIO("Change Name")

test_group = TestGroup()

def on_name_change(value):
    test_group.FirstName = "Steven"
    test_group.LastName = "Smith"

test_group.NameChange.set_callback=on_name_change

html_test_group =HMTLIOGroup(test_group)


doc = dominate.document(title='Test Page')
with doc.head:
    link(rel='stylesheet', href='https://www.w3schools.com/w3css/4/w3.css')
    link(rel='stylesheet', href="/static/easyio.css")
    script(type='text/javascript', src='https://ajax.googleapis.com/ajax/libs/jquery/3.4.1/jquery.min.js')
    script(type='text/javascript', src='static/easyio.js')
with doc.body:
    with div(cls="flex-container"):
        with div(cls="w3-card w3-margin w3-border w3-padding"):
            html_test_group.html_table(
                headers=["ITEM", "VALUE", "UNITS"],
                show_units=True)
        with div(cls="w3-card w3-margin w3-border w3-padding"):
            html_test_group.html(
                headers=["ITEM", "VALUE", "UNITS"],
                show_set=False, show_reset=False, show_units=True, by_lambda_each=lambda x : 'Bathroom' not in x.addr)
    with table():
        HTMLIO(test_group.BathroomBreaks).html_row(show_reset=True, show_set=True, show_value=False)
        HTMLIO(test_group.NameChange).html_row(show_set = True, show_value=False)





@app.route('/')
def hello_world():
    html = doc.render()
    return Response(html)


@app.route('/api/io/getall')
def get_all_io():
    test_group.Age = random.randint(40,50)
    d = html_test_group.dumps()
    return Response(d)

@app.route('/api/io/<key>/<value>', methods=['POST'])
def set_io(key: str, value: str):
    attr = test_group.get_attribute(key=key)
    attr.write(value)
    return jsonify(success=True)




if __name__ == '__main__':
   app.run(debug=True)


import unittest
from pyrediseasyio.io.io_group import IOGroup
from pyrediseasyio import BooleanIO, IntIO, FloatIO, StringIO
from assertpy import assert_that


class TestGroup(IOGroup):
    Bool1 = BooleanIO("Boolean 1", "Bool1", False)
    Bool2 = BooleanIO("Boolean 2", "Bool2", True)
    Int1 = IntIO("Integer 1", "Int1")
    Int2 = IntIO("Integer 2", "Int2", default=34)
    Float1 = FloatIO("Float 1", "Float1", default=1.2)

    def __init__(self):
        super().__init__(channel="TestChannel")


class TestGroup2(IOGroup):
    Bool1 = BooleanIO("Boolean 1", "Bool1", False)
    Bool2 = BooleanIO("Boolean 2", "Bool2", True)
    Int1 = IntIO("Integer 1", "Int1")
    Int2 = IntIO("Integer 2", "Int2", default=34)
    Float1 = FloatIO("Float 1", "Float1", default=1.2)
    Message1 = StringIO("String 1", "String1")

    def __init__(self):
        super().__init__(channel="TestChannel")


class TestGroup3(IOGroup):
    def __init__(self, db=2):
        self.Bool1 = BooleanIO("Boolean 1", default=False)
        self.Bool2 = BooleanIO("Boolean 2", "Bool2", True)
        self.Int1 = IntIO("Integer 1", "Int1")
        self.Int2 = IntIO("Integer 2", "Int2", default=34)
        self.Float1 = FloatIO("Float 1", "Float1", default=1.2)
        super().__init__(db=db)


class TestReadWrite(unittest.TestCase):

    def setUp(self):
        self.group = TestGroup()
        self.group2 = TestGroup2()
        self.group3 = TestGroup3()
        self.group.flush_db()
        self.group2.flush_db()
        self.group3.flush_db()

    def test_that_defaults_are_applied(self):
        group = self.group
        group3 = self.group3

        assert_that(group.Bool1).is_false()
        assert_that(group.Bool2).is_true()
        assert_that(group.Int1).is_equal_to(0)
        assert_that(group.Int2).is_equal_to(34)
        assert_that(group.Float1).is_equal_to(1.2)

        assert_that(group3.Bool1).is_false()
        assert_that(group3.Bool2).is_true()
        assert_that(group3.Int1).is_equal_to(0)
        assert_that(group3.Int2).is_equal_to(34)
        assert_that(group3.Float1).is_equal_to(1.2)

    def test_basic_read_write(self):
        group1 = self.group
        group2 = self.group2

        group4 = TestGroup3(db=2)
        group5 = TestGroup3(db=2)

        group1.Bool1 = True
        assert_that(group2.Bool1).is_true()
        group1.Bool1 = False
        assert_that(group2.Bool1).is_false()

        group1.Int1 = 100
        assert_that(group2.Int1).is_equal_to(100)
        group1.Int1 = -100
        assert_that(group2.Int1).is_equal_to(-100)

        group4.Float1 = 123.4
        assert_that(group5.Float1).is_equal_to(123.4)

        group4.Bool1 = True
        assert_that(group5.Bool1).is_true()
        group4.Bool1 = False
        assert_that(group5.Bool1).is_false()

        group4.Int1 = 100
        assert_that(group5.Int1).is_equal_to(100)
        group4.Int1 = -100
        assert_that(group5.Int1).is_equal_to(-100)

        group4.Float1 = 123.4
        assert_that(group5.Float1).is_equal_to(123.4)

    def test_same_name_different_database(self):
        group1 = self.group
        group2 = self.group2
        group3 = self.group3
        group4 = TestGroup3(db=5)

        group1.Int1 = 234
        assert_that(group2.Int1).is_equal_to(234)
        assert_that(group3.Int1).is_not_equal_to(234)

        group4.Float1 = 555
        assert_that(group4.Float1).is_equal_to(555)
        assert_that(group3.Float1).is_not_equal_to(555)

    def test_that_group_can_be_saved_to_memory(self):
        group = self.group
        group2 = self.group2
        j = group.dump("SomeKey")
        assert_that(j).is_not_none()
        assert_that(len(j)).is_greater_than(10)

        r = group2.read("SomeKey")
        assert_that(r).is_equal_to(j)

    def test_arithmatic_operations(self):
        group1 = self.group
        group3 = self.group3

        group1.Float1 = 10.0
        group3.Float1 = 20.0

        assert_that(group1.Float1 + group3.Float1).is_equal_to(30)
        assert_that(group1.Float1 * group3.Float1).is_equal_to(200)
        assert_that(group3.Float1 / group1.Float1).is_equal_to(2)
        assert_that(group3.Float1 - group1.Float1).is_equal_to(10)

    def test_that_pubsub_possible_on_io_group(self):
        group = self.group
        group2 = self.group2

        group2.Message1.publish("This is a test")

        msg = None
        for m in group.listen():
            msg = m
            break

        assert_that("This is a test" in msg[1]).is_true()


    def test_values_can_be_set_to_default(self):
        group1 = self.group
        group1.Bool1 = True
        group1.Float1 = 99

        assert_that(group1.Bool1).is_true()
        assert_that(group1.Float1).is_equal_to(99)

        group1.set_all_to_defaults()
        assert_that(group1.Bool1).is_false()
        assert_that(group1.Float1).is_equal_to(1.2)





import unittest
from pyrediseasyio.io.io_group import IOGroup
from pyrediseasyio import BooleanIO, IntIO, FloatIO
from assertpy import assert_that


class TestGroup(IOGroup):
    length = 5
    Bool1 = BooleanIO("Boolean 1", "Bool1", False)
    Bool2 = BooleanIO("Boolean 2", "Bool2", True)
    Int1 = IntIO("Integer 1", "Int1")
    Int2 = IntIO("Integer 2", "Int2", default=34)
    Float1 = FloatIO("Float 1", "Float1", default=1.2)



class TestGroup2(IOGroup):
    length = 5
    Bool1 = BooleanIO("Boolean 1")
    Bool2 = BooleanIO("Boolean 2")
    Int1 = IntIO("Integer 1", "Int1")
    Int2 = IntIO("Integer 2", "Int2", default=34)
    Float1 = FloatIO("Float 1", "MyName", default=1.2)
    Float2 = FloatIO("Float 2", "OtherAddr")


test_group = TestGroup()
test_group2 = TestGroup2(namespace="Test")


class TestAttributes(unittest.TestCase):

    def test_get_all_attributes(self):
        attrs = test_group.get_attributes()
        assert_that(len(attrs)).is_equal_to(TestGroup.length)

    def test_limit_attributes_by_names(self):
        attrs = test_group.get_attributes(by_names=['Bool1','Float1'])
        assert_that(len(attrs)).is_equal_to(2)

    def test_limit_attributes_by_type(self):
        attrs = test_group.get_attributes(by_type=[IntIO, FloatIO])
        assert_that(len(attrs)).is_equal_to(3)

    def test_limit_by_io_filter(self):
        attrs = test_group.get_attributes(by_lambda_each=lambda x: x.value == 1.2)
        assert_that(len(attrs)).is_equal_to(1)

    def test_limit_by_io_result(self):
        attrs = test_group.get_attributes(by_lambda_results=lambda x: x[-1:])
        assert_that(len(attrs)).is_equal_to(1)
        assert_that(attrs[0].value).is_equal_to(34)

    def test_namespace(self):
        assert_that(test_group2.Bool1.key).is_equal_to(f'TestBool1')
        assert_that(test_group2.Float1.key).is_equal_to('TestMyName')

    def test_can_get_single_attribute(self):
        name_attr = test_group2.get_attribute(name="Float2")
        addr_attr = test_group2.get_attribute(addr="OtherAddr")
        key_attr = test_group2.get_attribute(key='TestOtherAddr')

        attrs = [name_attr, addr_attr, key_attr]
        for attr in attrs:
            assert_that(attr).is_not_none()
            assert_that(attr).is_instance_of(FloatIO)
            assert_that(attr).name = "Float2"

        assert_that(test_group2.get_attribute(name="NotExists")).is_none()



import unittest
from pyrediseasyio.io.io_group import IOGroup
from pyrediseasyio.html.html_io_group import HMTLIOGroup
from pyrediseasyio import BooleanIO, SingleIO, FloatIO, StringIO
from assertpy import assert_that
from htmldiffer import utils as diff_utils
import json


class TestGroup(IOGroup):
    Bool1 = BooleanIO("Boolean 1", units="On/Off")
    Float1 = FloatIO("Float 1", default=1.23, units="furlongs")


test_group = TestGroup(namespace='Pin1')


def html_equals(html1, html2):
    lines1 = [line.strip() for line in diff_utils.html2list(html1)]
    lines1 = [line for line in lines1 if len(line) > 0]
    lines2 = [line.strip() for line in diff_utils.html2list(html2)]
    lines2 = [line for line in lines2 if len(line) > 0]
    diffs = [(line[0], line[1]) for line in zip(lines1, lines2) if line[0] != line[1]]
    return lines1 == lines2


class TestHTML(unittest.TestCase):

    def test_boolean_div_conversion(self):
        h = HMTLIOGroup(test_group).html(show_units=True).render()
        expected = '''
        <div class="easyio_container" id="Pin1Bool1_io_container">
          <div class="easyio_io" data-type="BooleanIO" id="Pin1Bool1_io">
            <div class="easyio_name">Boolean 1</div>
            <div class="easyio_value" data-addr="Bool1" data-namespace="Pin1" id="Pin1Bool1" onchange="OnEasyIOValueChange(event)">False</div>
            <div class="easyio_units">On/Off</div>
          </div>
          <div class="easyio_io" data-type="FloatIO" id="Pin1Float1_io">
            <div class="easyio_name">Float 1</div>
            <div class="easyio_value" data-addr="Float1" data-namespace="Pin1" id="Pin1Float1" onchange="OnEasyIOValueChange(event)">1.23</div>
            <div class="easyio_units">furlongs</div>
          </div>
        </div>
        '''
        assert_that(html_equals(h, expected)).is_true()

    def test_boolean_table_conversion(self):
        h = HMTLIOGroup(test_group).html_table(show_units=True).render()
        expected = '''
        <table class="easyio_container" id="Bool1_io_container">
          <tr class="easyio_io" data-type="BooleanIO" id="Pin1Bool1_io">
            <td class="easyio_name">Boolean 1</td>
            <td class="easyio_value" data-addr="Bool1" data-namespace="Pin1" id="Pin1Bool1" onchange="OnEasyIOValueChange(event)">False</td>
            <td class="easyio_units">On/Off</td>
          </tr>
          <tr class="easyio_io" data-type="FloatIO" id="Pin1Float1_io">
            <td class="easyio_name">Float 1</td>
            <td class="easyio_value" data-addr="Float1" data-namespace="Pin1" id="Pin1Float1" onchange="OnEasyIOValueChange(event)">1.23</td>
            <td class="easyio_units">furlongs</td>
          </tr>
        </table>
        '''
        assert_that(html_equals(h, expected)).is_true()


    def test_different_id(self):
        h = HMTLIOGroup(test_group).html_table(html_id="NewID", show_units=True).render()
        expected = '''
        <table class="easyio_container" id="NewID">
          <tr class="easyio_io" data-type="BooleanIO" id="Pin1Bool1_io">
            <td class="easyio_name">Boolean 1</td>
            <td class="easyio_value" data-addr="Bool1" data-namespace="Pin1" id="Pin1Bool1" onchange="OnEasyIOValueChange(event)">False</td>
            <td class="easyio_units">On/Off</td>
          </tr>
          <tr class="easyio_io" data-type="FloatIO" id="Pin1Float1_io">
            <td class="easyio_name">Float 1</td>
            <td class="easyio_value" data-addr="Float1" data-namespace="Pin1" id="Pin1Float1" onchange="OnEasyIOValueChange(event)">1.23</td>
            <td class="easyio_units">furlongs</td>
          </tr>
        </table>
        '''
        assert_that(html_equals(h, expected)).is_true()


    def test_set_reset_true(self):
        h = HMTLIOGroup(test_group).html_table(
            html_id="NewID", show_units=False, show_set=True, show_reset=True, set_text="SetMe").render()
        expected = '''
        <table class="easyio_container" id="NewID">
          <tr class="easyio_io" data-type="BooleanIO" id="Pin1Bool1_io">
            <td class="easyio_name">Boolean 1</td>
            <td class="easyio_value" data-addr="Bool1" data-namespace="Pin1" id="Pin1Bool1" onchange="OnEasyIOValueChange(event)">False</td>
            <td class="easyio_btn_cell easyio_set_btn_cell">
              <button class="easyio_set" onclick="EasyIOSet('Pin1','Bool1','Pin1Bool1',true)">SetMe</button>
            </td>
            <td class="easyio_btn_cell easyio_rst_btn_cell">
              <button class="easyio_reset" onclick="EasyIOSet('Pin1','Bool1','Pin1Bool1',true)">Reset</button>
            </td>
          </tr>
          <tr class="easyio_io" data-type="FloatIO" id="Pin1Float1_io">
            <td class="easyio_name">Float 1</td>
            <td class="easyio_value" data-addr="Float1" data-namespace="Pin1" id="Pin1Float1" onchange="OnEasyIOValueChange(event)">1.23</td>
            <td class="easyio_btn_cell easyio_set_btn_cell">
              <button class="easyio_set" onclick="EasyIOSet('Pin1','Float1','Pin1Float1',1.23)">SetMe</button>
            </td>
            <td class="easyio_btn_cell easyio_rst_btn_cell">
              <button class="easyio_reset" onclick="EasyIOSet('Pin1','Float1','Pin1Float1',0)">Reset</button>
            </td>
          </tr>
        </table>
        '''
        assert_that(html_equals(h, expected)).is_true()



    def test_json_dumps_basic(self):
        h = HMTLIOGroup(test_group)
        d = h.dumps()
        as_list = json.loads(d)
        first = as_list[0]
        assert_that(len(as_list)).is_equal_to(len(test_group))
        assert_that(first['name']).is_equal_to('Boolean 1')
        assert_that(first['display_value']).is_equal_to("Off")


    def test_json_dumps_by_name(self):
        h = HMTLIOGroup(test_group)
        d = h.dumps(by_names=['Bool1'])
        as_list = json.loads(d)
        assert_that(len(as_list)).is_equal_to(1)



from pyrediseasyio.io.numeric_io import NumericIO
from pyrediseasyio.io.single_io import SingleIO
from pyrediseasyio.abstract_reader_writer import AbstractReaderWriter


class FloatIO(NumericIO):
    def __init__(self, name: str, addr: str = None, default: float=0.0, **kwargs):
        """
        :param name:    A human readable name to give to the IO
        :param addr:    An optional address (key), if not given, them namespace + member name will be used
        :param default: The value to assign to the IO if no key is found during read
        :param kwargs:
         - units:       str: Optional units to assign to the IO
         - on_value:    str: The 'display_value' property will optionally return this value when the value is True
         - on_value:    str: The 'display_value' property will optionally return this value when the value is False
         - namespace:   str: Optional leading text to apply to the address to makes its key unique
        """
        super().__init__(name, addr, default, **kwargs)

    @staticmethod
    def _convert_type(value):
        return float(value)






import redis
from pyrediseasyio.abstract_reader_writer import AbstractReaderWriter
from typing import List


class ReaderWriter(AbstractReaderWriter):

    def __init__(self, host='localhost', port=6379, db=0,
                 channel: str = None, channels: List[str] = None):
        """
        :param host: If pointing to an external server, include the url here
        :param db: The db number can be used to segregate communication
        :param channel: for Pub/Sub only, the channel to listen to
        :param channels: For Pub/Sub only, the _channels to listen to
        Note, either supply channel, or _channels - not both.
        """
        self._server = redis.StrictRedis(host=host, port=port, db=db, decode_responses=True)
        self._pubsub = self._server.pubsub(ignore_subscribe_messages=True)

        # If _channels provided, then subscribe
        if channel is None and channels is None:
            return
        self._channels = []  # type: List[str]
        self._channels = [channel] if channel is not None else channels
        self._pubsub.subscribe(self._channels)

    def flush_db(self):
        self._server.flushdb()

    def flush_all(self):
        self._server.flushall()

    def delete_key(self, key: str):
        self._server.delete(key)

    @staticmethod
    def _get_channel_and_data_from_message(message: dict):
        c = message.get('channel')
        d = message.get('data')
        return c, d

    @property
    def channels(self) -> List[str]:
        """Names of all channels subscribed to for pub/sub"""
        return self._channels

    # Gets a message from the subscription if it exists, otherwise returns nothing.
    def get_next_message(self, get_subscription_message=False):
        """
        Checks to see if a single message is waiting and returns it, otherwise returns None
        :param get_subscription_message: If this is true, messages will also include first-time subscription
        notifications. It is generally better to leave this False as the data response won't match expectations.
        :return: tuple (channel, data) or None
        note: If multiple messages are waiting, only the first in line will be returned.
        """
        message = self._pubsub.get_message(ignore_subscribe_messages=not get_subscription_message)
        if message is None:
            return None
        return self._get_channel_and_data_from_message(message)

    def get_messages(self, include_subscription_messages=False, limit=100):
        """
        Checks for waiting messages and returns all of them, an empty list if none are waiting.
        :param include_subscription_messages: If this is true, messages will also include first-time subscription
        notifications. It is generally better to leave this False as the data response won't match expectations.
        :param limit: Applies a cap to the number of messages that will be returned.
        :return: Returns a list of tuples (channel, data) or an empty list if no messages are found.
        """
        messages = []
        msg = self.get_next_message(include_subscription_messages)
        while msg is not None:
            messages.append(msg)
            msg = self.get_next_message(include_subscription_messages)
            if len(messages) >= limit:
                break
        return messages

    def listen(self):
        """
        A blocking call to pub/sub - listens for a message and yields response if one arrives
        :return yields a tuple of (channel, data) under normal conditions. If error, yields tuple of ("Error", ex)
        """
        for message in self._pubsub.listen():
            try:
                channel, data = self._get_channel_and_data_from_message(message)
                yield channel, data
            except Exception as ex:
                yield 'Error', ex

    def publish(self, value: str, channel: str = None):
        """
        Publishes a message to all subscribers
        :param value: The value to publish, must be a string.
        :param channel: If none, will write to all _channels
        :return: Nothing
        """
        channels = [channel] if channel is not None else self._channels
        for channel in channels:
            self._server.publish(channel, value)

    def read(self, addr):
        return self._server.get(addr)

    def write(self, addr, value):
        value = str(value)
        self._server.set(addr, value)



